/* This program borns from an idea of Laszlo "rozsaheg" published in RocRail forum:
*  http://forum.rocrail.net/viewtopic.php?f=11&t=10120&start=15
*  http://forum.rocrail.net/download/file.php?id=22799
*/

/*   This program is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Version 01 modified by antogar on march 5 2017
* this version is suitable fot Arduino UNO, Nano etc with an easy configuration of pins
* there is a timer set to 500 ms for each pin for debouncing pourposes, useful for 
* point sensors based in IR barriers in order to avoid false detection with freight 
* trains with containers wagons
* 
* Version 02 modified by antogar april 2017
* from version 01, the status of the pin is sampled three times
* to avoid any false detection
* 
* Version 03 modified by antogar june 2017 
* from rev 02 added I2C chain management
* reduced the pin number to 15 due A4/A5 unavailability 
*  
* Version 04 modified by antogar agust 2017 
* from rev 03 modified I2C  management with PCF8574
* the I2C interface manages PCF8574 pin as inputs
* the 8 bit of PCF are sent to RR as single pin starting
* from base address set with A0/A1/A2 of  8574 
*  
* Version 05 modified by antogar agust 2017 
* from rev 04 modified I2C  management with PCF8574
* the I2C interface manages PCF8574 pin as inputs
* the 8 bit of PCF are sent to RR as single pin starting
* from address 16, each PCF pin is sequentially added
* starting from the lowest base address
* the commented rows are used for debug with local serial
*  
* Version 06 modified by antogar agust 2017 
* from rev 05, A0 is reserved for push button
* and adjusted pin according the board development
* 
* Version 07 modified by antogar 16/9/2017 
* added led status = on if PCF boards are connected
* 
* Version 08 modified by antogar 9/10/2017 
* added replay to query request of sensors status.
* The board replay to the command "start of the day"
* received at server starting with the status of the sensors
* 
*/

// pin SCL su A5 
// pin SDA su A4

#include <Wire.h>
#include <PCF8574.h>


#define num_sensors 14    // number of on board pin a sensors
#define num_pcf     5     // max number of PCF8574 connected

#define SET_UP     A0  //pin button 

int sense[num_sensors] = {A1, A2, A3, 12, 9, 8, 7, 6, 5, 4, 3, 2, 10, 11}; //  sensor configuration
byte pcf[num_pcf], prev_pcf[num_pcf];

int b=num_sensors;             // number of sensors
int sensors[num_sensors];      // status "matrix"

int i = 0, j = 0;
int chnum;
byte error, address;
int nDevices = 0;
boolean ledState = 0;
byte indirizzo[num_pcf];


unsigned long sensorprevmillis[num_sensors];
unsigned long pcfprevmillis[num_pcf][8];

int interval = 500;

byte tx_msg[13];      // CAN message to transmit
int bus_addr = 1;     // address of the feedback board chain


// Subroutine to send status to RR

void sendstatus (){

  for (int bp = 1; bp <= b; bp++) {
    sensors[bp-1] = digitalRead(sense[bp-1]);
    tx_msg[8]=bp;                       // sensor number
    tx_msg[10]=sensors[bp-1];           // pin status
    Serial.write(tx_msg,13);            // send the pin status to Rocrail
    delay(50);
    }

  if (nDevices != 0)
   {
    ledState = 1;
    for(i = 0; i < nDevices; i++ )
    {
      Wire.requestFrom(indirizzo[i], 1);
      while (Wire.available()) {
         byte c = Wire.read();
         pcf[i] = c;
         prev_pcf[i]=c;
         for (j = 0; j< 8; j++ ) {
            int value = bitRead(c,j);
            tx_msg[8]=(b + 1) + (i*8) +j;
            tx_msg[10]=value;
            Serial.write(tx_msg,13);
            delay(50);
         }
          
      }
     }
    }
}


// INIT

void setup() {
  Wire.begin();          // join i2c bus as master
 	Serial.begin(500000);

  pinMode(13, OUTPUT);
 
  pinMode(SET_UP, INPUT);
  digitalWrite(SET_UP, HIGH);

	tx_msg[0] = 0;      // prio			
	tx_msg[1] = 35;     // command - 0x11 command of MCS2 protokoll
	tx_msg[2] = 59;     // highByte(hash) - 0x3b
	tx_msg[3] = 121;    // lowByte(hash) - 0x79
	tx_msg[4] = 8;      // the number of data bytes in the payload			
	tx_msg[5] = 0;
	tx_msg[6] = bus_addr; // bus number
	tx_msg[7] = 0;
	tx_msg[9] = 0;			
	tx_msg[11]= 0;
	tx_msg[12]= 0;

  // Initialization of on board sensors pin
  
	for (int bp = 1; bp <= b; bp++) {
		pinMode(sense[bp-1], INPUT_PULLUP);
    sensors[bp-1] = digitalRead(sense[bp-1]);
    tx_msg[8]=bp;			                  // sensor number
		tx_msg[10]=sensors[bp-1];           // pin status
		Serial.write(tx_msg,13);	          // send the pin status to Rocrail
    delay(50);
	}

  // Scan i2c to find PCF8574s connected

  for(address = 1; address < 127; address++ ) 
  {
    Wire.beginTransmission(address);
    error = Wire.endTransmission();
    if (error == 0)
    {
//      Serial.print("I2C device found at address ");
//      if (address<16) Serial.print("0");
//      Serial.print(address);
//      Serial.println("  !");
      indirizzo[nDevices] = address;
      nDevices++;
    }
    else if (error==4) 
    {
//      Serial.print("Unknown error at address ");
//      if (address<16) 
//        Serial.print("0");
//      Serial.println(address);
    }    
  }


  // If there are PCF8574s connected, send the status to RR
  
  if (nDevices != 0)
   {
    ledState = 1;
    for(i = 0; i < nDevices; i++ )
    {
      Wire.requestFrom(indirizzo[i], 1);
      while (Wire.available()) {
         byte c = Wire.read();
         pcf[i] = c;
         prev_pcf[i]=c;
///         Serial.println(pcf[i]);
         for (j = 0; j< 8; j++ ) {
            int value = bitRead(c,j);
            tx_msg[8]=(b + 1) + (i*8) +j;
            tx_msg[10]=value;
//            Serial.print(tx_msg[8]);
//            Serial.print("-");
//            Serial.println(tx_msg[10]);
              Serial.write(tx_msg,13);
              delay(50);
         }
          
      }

     }
    }
 digitalWrite(13, ledState);

}

void loop() {

if (Serial.available()) {
    byte ch = Serial.read();
    if (ch==34 && chnum==0 )
      chnum++;
      else if (ch==19 && chnum==1 )
        chnum++;
        else if (ch==20 && chnum==2 )
          {
            chnum = 0;
            sendstatus ();
//            PortaSeriale.println(" Richiesta Polling! ");
          }
          else
            chnum=0;
//    PortaSeriale.print(ch);
//    PortaSeriale.print(" ");
//    PortaSeriale.println(chnum);    
    }


  // Scan the on board pin
  
  for (int bp = 1; bp <= b; bp++) {	  		
  	if ((millis() - sensorprevmillis [bp-1]) > interval)      // is the interval expired for this sensor ? there was change on pin state compare to previous scan ?
    	if (digitalRead(sense[bp-1])!=sensors[bp-1])            // if yes sample the pin three times with 10ms interval
    	    {                                                       
    			delay (10);
          if (digitalRead(sense[bp-1])!=sensors[bp-1])
            {
            delay (10);
            if (digitalRead(sense[bp-1])!=sensors[bp-1])
                    {
              			sensorprevmillis [bp-1] = millis();                 // if yes, change the bit and reset the sensor timer
              			sensors[bp-1] = digitalRead(sense[bp-1]);           // store the current status in 'sensors' variable
              			tx_msg[8]=bp;			                                  // the sensor number is sensor address
              			tx_msg[10]=sensors[bp-1];                           // set pin status
//                    Serial.print(tx_msg[8]);
//                    Serial.print("-");
//                    Serial.println(tx_msg[10]);
              			Serial.write(tx_msg,13);		                        // send the message status to Rocrail		
                    }
            }

    	    }
    }

  // If there are PCF8574s connected, scan the status
  
  if (nDevices != 0)
   {
    for(i = 0; i < nDevices; i++ )
    {
      for(j = 0; j < 8; j++ )
      {
        if ((millis() - pcfprevmillis [i][j]) > interval)
        {
          
          Wire.requestFrom(indirizzo[i], 1);
          pcf[i] = Wire.read();
          if (bitRead(pcf[i],j) != bitRead(prev_pcf[i],j))
          {
            delay(10);
            Wire.requestFrom(indirizzo[i], 1);
            pcf[i] = Wire.read();
            if (bitRead(pcf[i],j) != bitRead(prev_pcf[i],j))
            {
              delay(10);
              Wire.requestFrom(indirizzo[i], 1);
              pcf[i] = Wire.read();
              if (bitRead(pcf[i],j) != bitRead(prev_pcf[i],j))
              {
                pcfprevmillis [i][j] = millis();
                bitWrite(prev_pcf[i],j,bitRead(pcf[i],j));
                tx_msg[8]= (b + 1) + (i*8) +j;
                tx_msg[10]=bitRead(pcf[i],j);
//                Serial.print(tx_msg[8]);
//                Serial.print("-");
//                Serial.println(tx_msg[10]);
                  Serial.write(tx_msg,13);
               }
             }
           }
        }
        
      }     
    } 
   }


}


